/*
 * virthost.h
 *
 * Utility routines that interface to the serial stream hardware
 */

#ifndef INC_VIRTHOST_H_
#define INC_VIRTHOST_H_

#include <stdint.h>
#include <stdbool.h>


//
// Constants
//

// ASCII Characters
#define NUL     0x00
#define EOT     0x04
#define CRET    '\r'
#define NL      '\n'
#define ENQ     0x05
#define ACK     0x06
#define BS      0x08
#define LF      0x0a
#define VT      0x0b
#define FF      0x0c
#define TAB     '\t'
#define BELL    0x07
#define SO      0x0e
#define SI      0x0f
#define DC1     0x11
#define DC2     0x12
#define DC3     0x13
#define DC4     0x14
#define NACK    0x15
#define XON     0x11
#define XOFF    0x13
#define ETB     0x17
#define CANCEL  0x18
#define SUB     0x1a
#define ESC     0x1b
#define SPACE   ' '
#define SQUOTE  '\''
#define DQUOTE  '\"'
#define CTRLC   0x03
#define CTRLH   0x08
#define CTRLS   0x13
#define CTRLX   0x18
#define DEL     0x7f

// Simulated slow print baud rate
#define VH_BAUD 300


//
// API
//
bool vh_InterfaceAvailable();
bool vh_InputAvailable();
char vh_InChar();

void vh_ClearScreen();
void vh_OutChar(char c);
void vh_PrintPrompt(const char* m);
void vh_PrintString(const char* m);
void vh_FastPrintString(const char* m);
void vh_OutFlush();


#endif /* INC_VIRTHOST_H_ */
