/*
 * virthost.c
 *
 * Utility routines that interface to the serial stream hardware
 */
#include "virthost.h"
#include "usbd_cdc_if.h"
#include <stdio.h>



// ===========================================================================================
// Local constants
// ===========================================================================================
#define SLOW_DELAY_MSEC (10 * 1000 / VH_BAUD)


// ===========================================================================================
// Local variables
// ===========================================================================================
static bool slow_print;



// ===========================================================================================
// API
// ===========================================================================================

bool vh_InterfaceAvailable()
{
	return usbd_cdc_is_connected;
}


bool vh_InputAvailable()
{
	return (usbd_cdc_rx_push_index != usbd_cdc_rx_pop_index);
}


char vh_InChar()
{
	char c;

	c = UserRxBufferFS[usbd_cdc_rx_pop_index++];
	if (usbd_cdc_rx_pop_index == APP_RX_DATA_SIZE) usbd_cdc_rx_pop_index = 0;

	return c;
}


void vh_ClearScreen()
{
	// Do this fast
	slow_print = false;

	// Home and Clear screen (ANSI)
	vh_OutChar(ESC);
	vh_OutChar('[');
	vh_OutChar('H');
	vh_OutChar(ESC);
	vh_OutChar('[');
	vh_OutChar('J');

	// Add NL + CR for people not running an ANSI terminal emulator
	vh_OutChar(NL);
	vh_OutChar(CRET);
	vh_OutFlush();
}


void vh_OutChar(char c)
{
	if (usbd_cdc_tx_len < APP_TX_DATA_SIZE) {
		UserTxBufferFS[usbd_cdc_tx_len++] = (uint8_t) c;
	}

	if (slow_print) {
		vh_OutFlush();
		HAL_Delay(SLOW_DELAY_MSEC);
	}
}


void vh_PrintPrompt(const char* m)
{
	slow_print = true;

	while (*m != 0) {
		vh_OutChar((char) *m++);
	}
	vh_OutFlush();
}


void vh_PrintString(const char* m)
{
	slow_print = true;

	while (*m != 0) {
		vh_OutChar((char) *m++);
	}
	vh_OutChar(NL);
	vh_OutChar(CRET);
	vh_OutFlush();
}


void vh_FastPrintString(const char* m)
{
	slow_print = false;

	while (*m != 0) {
		vh_OutChar((char) *m++);
	}
	vh_OutChar(NL);
	vh_OutChar(CRET);
	vh_OutFlush();
}


void vh_OutFlush()
{
	while (CDC_Transmit_FS(UserTxBufferFS, usbd_cdc_tx_len) == USBD_BUSY) {
		HAL_Delay(1);
	}
	usbd_cdc_tx_len = 0;
}
